# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import timedelta
from urllib.parse import quote


class BaseAssetSigner:
    def __init__(self, public=False):
        self.public = public

    @property
    def signature_expiry_duration(self) -> timedelta:
        return timedelta(minutes=15)

    @property
    def signature_required(self) -> bool:
        return not self.public

    # espect subclass may have some costly checking in this method
    def available(self) -> bool:
        return True

    def percent_escape_asset_name(self, name: str) -> str:
        # We do not quote '/' in name.
        # The first reason is that most file system
        # does not allow '/' be part of the filename.
        # Another reason is that AWS S3 allows object
        # key to contain '/'.
        return quote(name, safe='/')



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import json
import logging
import os
from functools import wraps

from werkzeug.test import EnvironBuilder
from werkzeug.wrappers import BaseResponse, Request

from ..encoding import (_serialize_exc, deserialize_or_none, deserialize_value,
                        serialize_record, serialize_value)
from ..error import SkygearException
from ..registry import get_registry
from ..utils import db
from ..utils.context import start_context
from ..utils.logging import setLoggerTag

log = logging.getLogger(__name__)
setLoggerTag(log, 'plugin')


def _get_engine():
    return db._get_engine()


def _wrap_result(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        try:
            return dict(result=f(self, *args, **kwargs))
        except Exception as e:
            handler = get_registry().get_exception_handler(e.__class__)
            if not handler:
                handler = handle_exception
            result = handler(e)
            if result is None:
                return dict(error=_serialize_exc(e).as_dict())
            elif isinstance(result, Exception):
                return dict(error=_serialize_exc(result).as_dict())
            else:
                return result
    return wrapper


def handle_exception(exc):
    if not isinstance(exc, SkygearException):
        log.exception("Error occurred processing request")
    return exc


def handle_init_event(**data):
    return get_registry().func_list()


def encode_base64_json(data):
    """
    Encode dict-like data into a base64 encoded JSON string.

    This can be used to get dict-like data into HTTP headers / envvar.
    """
    return base64.b64encode(bytes(json.dumps(data), 'utf-8'))


def decode_base64_json(data):
    """
    Decode dict-like data from a base64 encoded JSON string.

    This can be used to get dict-like data into HTTP headers / envvar.
    """
    return json.loads(base64.b64decode(data).decode('utf-8'))


def dict_from_base64_environ(name):
    data = os.environ.get(name)
    return decode_base64_json(data) if data else {}


class CommonTransport:
    def __init__(self, registry=None):
        self._registry = registry or get_registry()

    @_wrap_result
    def call_func(self, ctx, kind, name, param):
        obj = self._registry.get_func(kind, name)
        with start_context(ctx):
            if kind == 'op':
                return self.op(obj, param.get('args', {}))
            elif kind == 'hook':
                return self.hook(obj, param)
            elif kind == 'timer':
                return self.timer(obj)
            else:
                raise SkygearException("unknown plugin extension point")

    @_wrap_result
    def call_event_func(self, name, param):
        try:
            event_funcs = self._registry.get_event_funcs(name)
        except KeyError:
            log.warning('Missing event func named "{}"'.format(name))
            return

        if name == 'init':
            # Only init event support returning data
            return self.event(event_funcs[0], param)
        else:
            for event_func in event_funcs:
                self.event(event_func, param)

    @_wrap_result
    def call_provider(self, ctx, name, action, param):
        obj = self._registry.get_provider(name)

        with start_context(ctx):
            return self.provider(obj, action, param)

    @_wrap_result
    def call_handler(self, ctx, name, param):
        func = self._registry.get_handler(name, param['method'])
        with start_context(ctx):
            return self.handler(func, param)

    def handler(self, func, param):
        builder = EnvironBuilder(
            method=param['method'],
            path=param['path'],
            query_string=param.get('query_string'),
            headers=param['header'],
            data=base64.b64decode(param['body'])
        )
        environ = builder.get_environ()
        request = Request(environ, populate_request=False, shallow=False)
        response = func(request)
        status = 200
        if isinstance(response, BaseResponse):
            headers = {}
            for k, v in response.headers:
                headers[k] = [v]
            body_byte = response.get_data()
            body = base64.b64encode(body_byte).decode('utf-8')
            status = response.status_code
        elif isinstance(response, str):
            headers = {'Content-Type': ['text/plain; charset=utf-8']}
            body = base64.b64encode(
                bytes(response, 'utf-8')
            ).decode('utf-8')
        else:
            headers = {
                'Content-Type': ['application/json']
             }
            body = encode_base64_json(response).decode('utf-8')
        return {
            'status': status,
            'header': headers,
            'body': body
        }

    def op(self, func, param):
        if isinstance(param, list):
            args = deserialize_value(param)
            kwargs = {}
        elif isinstance(param, dict):
            args = []
            kwargs = deserialize_value(param)
        else:
            msg = "Unsupported args type '{0}'".format(type(param))
            raise ValueError(msg)
        return serialize_value(func(*args, **kwargs))

    def hook(self, func, param):
        original_record = deserialize_or_none(param.get('original', None))
        record = deserialize_or_none(param.get('record', None))
        with db.conn() as conn:
            returned = func(record, original_record, conn)

            # If the hook function does not return a value, assume that
            # the record in the first argument is to be returned.
            if returned is None:
                returned = record
        return serialize_record(returned)

    def timer(self, func):
        return func()

    def event(self, func, data):
        if not isinstance(data, dict):
            msg = "Unsupported args type '{0}'".format(type(data))
            raise ValueError(msg)
        return func(**data)

    def provider(self, provider, action, data):
        return provider.handle_action(action, data)

    def run(self):
        raise NotImplemented()


_registry = get_registry()
_registry.register_exception_handler(Exception, handle_exception)
_registry.register_event('init', handle_init_event)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import logging

import requests
import strict_rfc3339

from .__version__ import __version__
from .database import Database
from .utils.logging import setLoggerTag

log = logging.getLogger(__name__)
setLoggerTag(log, 'plugin')


class PayloadEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            ts = obj.timestamp()
            return strict_rfc3339.timestamp_to_rfc3339_utcoffset(ts)


def send_action(url, payload, timeout=60):
    log.error("skygear.container.send_action is deprecated.\n"
              "Please use SkygearContainer().send_action instead.")
    headers = {'Content-type': 'application/json',
               'Accept': 'application/json',
               'X-Skygear-SDK-Version': 'py-skygear/' + __version__}
    _data = json.dumps(payload, cls=PayloadEncoder)
    return requests.post(url, data=_data, headers=headers, timeout=timeout) \
        .json()


class SkygearContainer(object):
    endpoint = 'http://localhost:3000'
    """Skygear Server Location"""
    api_key = None
    access_token = None
    user_id = None
    app_name = ''
    transport = None

    def __init__(self, endpoint=None, api_key=None, access_token=None,
                 user_id=None, transport=None):
        """
        `public_database` Public Database
        `private_database` Private Database
        """
        if endpoint:
            self.endpoint = endpoint
        if api_key:
            self.api_key = api_key
        if user_id:
            self.user_id = user_id
        self.access_token = access_token
        if transport is None:
            self.transport = SkygearContainer.transport
        else:
            self.transport = transport
        self.public_database = Database(self, '_public')
        self.private_database = Database(self, '_private')

    def _request_url(self, action_name):
        endpoint = self.endpoint
        endpoint = endpoint[:-1] if endpoint[-1] == '/' else endpoint
        return endpoint + '/' + action_name.replace(':', '/')

    def _payload(self, action_name, params, plugin_request):
        payload = params.copy() if isinstance(params, dict) else {}
        payload['action'] = action_name
        if self.access_token:
            payload['access_token'] = self.access_token
        if self.api_key:
            payload['api_key'] = self.api_key
        if self.user_id:
            payload['_user_id'] = self.user_id
        if plugin_request:
            payload['_from_plugin'] = True
        return payload

    @classmethod
    def set_default_app_name(cls, app_name):
        cls.app_name = app_name

    @classmethod
    def get_default_app_name(cls):
        return cls.app_name

    @classmethod
    def set_default_endpoint(cls, endpoint):
        cls.endpoint = endpoint

    @classmethod
    def set_default_apikey(cls, api_key):
        cls.api_key = api_key

    @classmethod
    def set_default_transport(cls, transport):
        cls.transport = transport

    def send_action(self, action_name, params, plugin_request=False,
                    timeout=60):
        url = self._request_url(action_name)
        payload = self._payload(action_name, params, plugin_request)
        resp = self.transport.send_action(action_name, payload, url, timeout)
        return resp



# Copyright 2017 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .asset import get_signer
from .encoding import deserialize_record, serialize_record
from .error import RecordQueryInvalid, SkygearException
from .models import Record


class Database(object):
    """The Skygear database CRUD class.

    Example:

    >>> database = container.public_database
    >>> database.save(record)

    """

    def __init__(self, container, database_id):
        self.container = container
        self.database_id = database_id

    def save(self, arg, atomic=False):
        """Save Function.

        Args:
            arg1 (list): A list of records
            atomic (bool): Atomic save if true. Defaults to False

        Returns:
            dict: Skygear server response

        Raises:
            SkygearException: If skygear server returns an error.
        """

        if not isinstance(arg, list):
            arg = [arg]
        records = [serialize_record(item)
                   if isinstance(item, Record) else item
                   for item in arg]
        return self.container.send_action('record:save', {
            'database_id': self.database_id,
            'records': records,
            'atomic': atomic
        })

    @staticmethod
    def _encode_id(record_id):
        return record_id.type + "/" + record_id.key

    def delete(self, arg):
        """
        Delete records.


        Args:
            arg (list): List of records or ID

        Returns:
            dict: Skygear server response

        Raises:
            SkygearException: If skygear server returns an error.
        """
        if not isinstance(arg, list):
            arg = [arg]
        ids = [Database._encode_id(item.id)
               if isinstance(item, Record)
               else item
               for item in arg]
        return self.container.send_action('record:delete', {
            'database_id': self.database_id,
            'ids': ids
        })

    def query(self, query):
        """Query records.

        Args:
            query (Query): Query object

        Returns:
            list: List of Record

        Raises:
            SkygearException: If skygear server returns an error.
        """

        include = {v: {"$type": "keypath", "$val": v}
                   for v in list(set(query.include))}

        payload = {'database_id': self.database_id,
                   'record_type': query.record_type,
                   'predicate': query.predicate.to_dict(),
                   'count': query.count,
                   'sort': query.sort,
                   'include': include}

        if query.offset is not None:
            payload['offset'] = query.offset
        if query.limit is not None:
            payload['limit'] = query.limit
        result = self.container.send_action('record:query', payload)
        if 'error' in result:
            raise SkygearException(result['error']['message'],
                                   code=RecordQueryInvalid)
        result = result['result']
        output = []
        signer = None
        for r in result:
            record = deserialize_record(r)
            if '_transient' in r:
                t = r['_transient']
                record['_transient'] = {k: deserialize_record(t[k])
                                        for k in t.keys()}
            if 'attachment' in r:
                if signer is None:
                    signer = get_signer()
                record['attachment'] = r['attachment'].copy()
                record['attachment']['$url'] =\
                    signer.sign(r['attachment']['$name'])
            output.append(record)
        return output



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from skygear.models import (Asset, Location, PublicAccessControlEntry, Record,
                            RecordID, Reference, RoleAccessControlEntry,
                            UnknownValue)

from ...encoding import (deserialize_record, deserialize_value,
                         serialize_record, serialize_value)


class TestsDeserializeRecord():
    def test_normal(self):
        rdata = {
            "_access": None,
            "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
            "_ownerID": "OWNER_ID",
            "content": "Hello World!",
            "noteOrder": 1,
            "money": {"$type": "unknown", "$underlying_type": "money"},
            }
        r = deserialize_record(rdata)
        assert r.id.type == "note"
        assert r.id.key == "99D92DBA-74D5-477F-B35E-F735E21B2DD5"
        assert r['content'] == "Hello World!"
        assert r['noteOrder'] == 1
        assert r.owner_id == "OWNER_ID"
        assert r['money'].underlyingType == "money"

    def test_model_with_acl(self):
        rdata = {
            "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
            "_ownerID": "OWNER _ID",
            "_access": [{
                "role": "admin",
                "level": "write"
            }]
        }
        r = deserialize_record(rdata)
        assert r.id.type == "note"
        assert r.id.key == "99D92DBA-74D5-477F-B35E-F735E21B2DD5"
        assert isinstance(r.acl, list)
        assert isinstance(r.acl[0], RoleAccessControlEntry)

    def test_model_with_public_read_acl(self):
        rdata = {
            "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
            "_ownerID": "OWNER _ID",
            "_access": [{
                "public": True,
                "level": "read"
            }]
        }
        r = deserialize_record(rdata)
        assert r.id.type == "note"
        assert r.id.key == "99D92DBA-74D5-477F-B35E-F735E21B2DD5"
        assert isinstance(r.acl, list)
        assert isinstance(r.acl[0], PublicAccessControlEntry)
        assert r.acl[0].level == "read"

    def test_serialize_record_back(self):
        rdata = {
            "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
            "_ownerID": "OWNER_ID",
            "_access": [{
                "public": True,
                "level": "read"
            }],
            "_created_at": "2014-09-27T17:40:00Z"
        }
        r = deserialize_record(rdata)
        result = serialize_record(r)
        assert result['_id'] == "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5"
        assert result['_ownerID'] == "OWNER_ID"
        assert result['_access'] == [{
            "public": True,
            "level": "read"
        }]
        assert result['_created_at'] == "2014-09-27T17:40:00Z"


class TestsSerializeRecord():
    def test_normal(self):
        r = Record(
            id=RecordID('note', '99D92DBA-74D5-477F-B35E-F735E21B2DD5'),
            owner_id='OWNER_ID',
            acl=None,
            data={
                  "content": "Hello World!",
                  "noteOrder": 1,
                  "money": UnknownValue("money"),
                  })
        rdata = serialize_record(r)
        assert rdata == {
            "_access": None,
            "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
            "_ownerID": "OWNER_ID",
            "content": "Hello World!",
            "noteOrder": 1,
            "money": {'$type': 'unknown', '$underlying_type': 'money'},
        }


class TestsDeserializeValue():
    def test_integer(self):
        assert deserialize_value(42) == 42

    def test_null(self):
        assert deserialize_value(None) is None

    def test_boolean(self):
        assert deserialize_value(True) is True

    def test_date(self):
        value = {"$type": "date", "$date": "2016-12-08T08:38:33Z"}
        assert deserialize_value(value) == datetime.fromtimestamp(1481186313)

    def test_asset(self):
        value = {
            "$type": "asset",
            "$name": "c1d0e8d4-648c-4c88-86c6-22feb1a6e734",
            "$content_type": "text/html"
        }
        assert deserialize_value(value) == \
            Asset("c1d0e8d4-648c-4c88-86c6-22feb1a6e734", "text/html")

    def test_location(self):
        value = {"$type": "geo", "$lng": 1, "$lat": 2}
        assert deserialize_value(value) == Location(1, 2)

    def test_reference(self):
        value = {
            "$type": "ref",
            "$id": "note/c1d0e8d4-648c-4c88-86c6-22feb1a6e734"
        }
        assert deserialize_value(value) == \
            Reference(RecordID("note", "c1d0e8d4-648c-4c88-86c6-22feb1a6e734"))

    def test_unknown_value(self):
        value = {"$type": "unknown", "$underlying_type": "unknown_type"}
        assert deserialize_value(value) == UnknownValue("unknown_type")

    def test_record(self):
        value = {
            "$type": "record",
            "$record": {
                "_access": None,
                "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
                "_ownerID": "OWNER_ID",
                "content": "Hello World!",
            }
        }
        r = deserialize_value(value)
        assert r.id.type == "note"
        assert r.id.key == "99D92DBA-74D5-477F-B35E-F735E21B2DD5"
        assert r.owner_id == "OWNER_ID"
        assert r["content"] == "Hello World!"

    def test_list(self):
        value = [42, {"$type": "geo", "$lng": 1, "$lat": 2}]
        assert deserialize_value(value) == [42, Location(1, 2)]

    def test_dict(self):
        value = {
            "distance": 42,
            "location": {"$type": "geo", "$lng": 1, "$lat": 2}
        }
        assert deserialize_value(value) == {
            "distance": 42,
            "location": Location(1, 2)
        }


class TestsSerializeValue():
    def test_integer(self):
        assert serialize_value(42) == 42

    def test_null(self):
        assert serialize_value(None) is None

    def test_boolean(self):
        assert serialize_value(True) is True

    def test_date(self):
        value = datetime.fromtimestamp(1481186313)
        assert serialize_value(value) == \
            {"$type": "date", "$date": "2016-12-08T08:38:33Z"}

    def test_asset(self):
        value = Asset("c1d0e8d4-648c-4c88-86c6-22feb1a6e734", "text/html")
        assert serialize_value(value) == \
            {
                "$type": "asset",
                "$name": "c1d0e8d4-648c-4c88-86c6-22feb1a6e734",
                "$content_type": "text/html"
            }

    def test_location(self):
        value = Location(1, 2)
        assert serialize_value(value) == \
            {"$type": "geo", "$lng": 1, "$lat": 2}

    def test_reference(self):
        value = Reference(
            RecordID("note", "c1d0e8d4-648c-4c88-86c6-22feb1a6e734"))
        assert serialize_value(value) == \
            {
                "$type": "ref",
                "$id": "note/c1d0e8d4-648c-4c88-86c6-22feb1a6e734"
            }

    def test_unknown_value(self):
        value = UnknownValue("unknown_type")
        assert serialize_value(value) == \
            {"$type": "unknown", "$underlying_type": "unknown_type"}

    def test_record(self):
        value = Record(
                id=RecordID("note", "99D92DBA-74D5-477F-B35E-F735E21B2DD5"),
                owner_id="OWNER_ID",
                acl=None,
                data={
                      "content": "Hello World!",
                      })
        assert serialize_value(value) == \
            {
                "$type": "record",
                "$record": {
                    "_access": None,
                    "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
                    "_ownerID": "OWNER_ID",
                    "content": "Hello World!",
                }
            }

    def test_list(self):
        value = [42, Location(1, 2)]
        assert serialize_value(value) == \
            [42, {"$type": "geo", "$lng": 1, "$lat": 2}]

    def test_dict(self):
        value = {
            "distance": 42,
            "location": Location(1, 2)
        }
        assert serialize_value(value) == \
            {
                "distance": 42,
                "location": {"$type": "geo", "$lng": 1, "$lat": 2}
            }



import unittest

from ..query import Predicate


class TestPredicate(unittest.TestCase):
    def test_simple_in(self):
        p = Predicate(_id__in=["a", "b", "c"])
        self.assertListEqual(
            ["in",
             {"$type": "keypath", "$val": "_id"},
             ["a", "b", "c"]],
            p.to_dict())

    def test_simple_not(self):
        p = ~Predicate(_id__in=["a", "b", "c"])
        self.assertListEqual(
            ['not',
             ["in",
              {"$type": "keypath", "$val": "_id"},
              ["a", "b", "c"]]],
            p.to_dict())

    def test_simple_and(self):
        p = Predicate(_id__eq="a", deleted__eq=False)
        expected = ["and",
                    ["eq",
                     {"$type": "keypath", "$val": "_id"},
                     "a"],
                    ["eq",
                     {"$type": "keypath", "$val": "deleted"},
                     False]]
        self.assertListEqual(
            expected,
            p.to_dict()
        )
        p = Predicate(_id__eq="a")
        p = p & Predicate(deleted__eq=False)
        self.assertListEqual(expected, p.to_dict())

    def test_simple_and_three_statements(self):
        p = Predicate(time__lte="2010-07-10",
                      time__gte="2009-01-01",
                      deleted__ne=False)
        expected = ["and",
                    ["ne",
                     {"$type": "keypath", "$val": "deleted"},
                     False],
                    ["gte",
                     {"$type": "keypath", "$val": "time"},
                     "2009-01-01"],
                    ["lte",
                     {"$type": "keypath", "$val": "time"},
                     "2010-07-10"]]
        self.assertListEqual(expected, p.to_dict())

    def test_simple_or(self):
        p = Predicate(_id__eq="simple", gender__eq="M", op=Predicate.OR)
        expected = ["or",
                    ["eq",
                     {"$type": "keypath", "$val": "_id"},
                     "simple"],
                    ["eq",
                     {"$type": "keypath", "$val": "gender"},
                     "M"]]
        self.assertListEqual(expected, p.to_dict())

    def test_simple_or_three_statement(self):
        p = Predicate(_id__eq="chima",
                      gender__eq="M",
                      type__in=["cat", "dog"],
                      op=Predicate.OR)
        expected = ["or",
                    ["eq",
                     {"$type": "keypath", "$val": "_id"},
                     "chima"],
                    ["eq",
                     {"$type": "keypath", "$val": "gender"},
                     "M"],
                    ["in",
                     {"$type": "keypath", "$val": "type"},
                     ["cat", "dog"]]]
        self.assertListEqual(expected, p.to_dict())

    def test_compound_statement_1(self):
        p = Predicate(_id__eq="chima", gender__eq="M", type__eq="dog")
        p2 = Predicate(_id__eq="fatseng", gender__eq="F", type__eq="cat")
        p3 = Predicate(_id__eq="milktea", gender__eq="NA", type__eq="frog")
        p4 = p & p2 & p3
        expected = ["and",
                    ["eq",
                     {"$type": "keypath", "$val": "_id"},
                     "chima"],
                    ["eq",
                     {"$type": "keypath", "$val": "gender"},
                     "M"],
                    ["eq",
                     {"$type": "keypath", "$val": "type"},
                     "dog"],
                    ["eq",
                     {"$type": "keypath", "$val": "_id"},
                     "fatseng"],
                    ["eq",
                     {"$type": "keypath", "$val": "gender"},
                     "F"],
                    ["eq",
                     {"$type": "keypath", "$val": "type"},
                     "cat"],
                    ["eq",
                     {"$type": "keypath", "$val": "_id"},
                     "milktea"],
                    ["eq",
                     {"$type": "keypath", "$val": "gender"},
                     "NA"],
                    ["eq",
                     {"$type": "keypath", "$val": "type"},
                     "frog"]]
        self.assertListEqual(expected, p4.to_dict())

    def test_compound_statement_2(self):
        p = Predicate(_id__eq="chima", gender__eq="M", type__eq="dog")
        p2 = Predicate(_id__eq="fatseng", gender__eq="F", type__eq="cat")
        p3 = ~Predicate(_id__eq="milktea", gender__eq="NA", type__eq="frog")
        p4 = p | p2 | p3
        expected_p = ["and",
                      ["eq", {"$type": "keypath", "$val": "_id"}, "chima"],
                      ["eq", {"$type": "keypath", "$val": "gender"}, "M"],
                      ["eq", {"$type": "keypath", "$val": "type"}, "dog"]]
        expected_p2 = ["and",
                       ["eq", {"$type": "keypath", "$val": "_id"}, "fatseng"],
                       ["eq", {"$type": "keypath", "$val": "gender"}, "F"],
                       ["eq", {"$type": "keypath", "$val": "type"}, "cat"]]
        expected_p3 = ["not",
                       ["and",
                        ["eq",
                         {"$type": "keypath", "$val": "_id"},
                         "milktea"],
                        ["eq", {"$type": "keypath", "$val": "gender"}, "NA"],
                        ["eq", {"$type": "keypath", "$val": "type"}, "frog"]]]
        expected = ["or", expected_p, expected_p2, expected_p3]
        self.assertListEqual(expected, p4.to_dict())



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import os
import unittest
from unittest.mock import MagicMock, patch

from .. import common
from ...encoding import deserialize_or_none, serialize_record
from ...error import SkygearException
from ...models import Record, RecordID
from ...registry import Registry
from ...utils.context import current_context
from ..common import CommonTransport


class TestCommonTransport(unittest.TestCase):
    def setUp(self):
        self.registry = Registry()
        self.transport = CommonTransport(self.registry)
        self.ctx = {'state': 'happy'}

    def tearDown(self):
        self.registry = None
        self.transport = None

    @patch('skygear.registry.Registry.get_func')
    def testCallFuncGetCorrectObject(self, mocker):
        mocker.return_value = MagicMock()
        self.transport.call_func(self.ctx, 'timer', 'name', {})
        mocker.assert_called_once_with('timer', 'name')
        mocker.return_value.assert_called_once_with()

    @patch('skygear.registry.Registry.get_func')
    def testCallFuncContext(self, mocker):
        def assertState(expectedState):
            def func():
                assert current_context().get('state') == expectedState
            return func

        mocker.return_value = MagicMock(side_effect=assertState("happy"))
        self.transport.call_func(self.ctx, 'timer', 'name', {})
        mocker.return_value.assert_called_once_with()
        assert 'state' not in current_context()

    @patch('skygear.registry.Registry.get_func')
    def testCallFuncResult(self, mocker):
        mocker.return_value = MagicMock(return_value={'data': 'hello'})
        result = self.transport.call_func(self.ctx, 'timer', 'name', {})
        assert result['result'] == {'data': 'hello'}

    @patch('skygear.registry.Registry.get_provider')
    def testCallProviderGetCorrectObject(self, mocker):
        mocker.return_value = MagicMock()
        self.transport.call_provider(self.ctx, 'name', 'action', {})
        mocker.assert_called_once_with('name')
        mocker.return_value.handle_action.assert_called_once_with('action', {})

    @patch('skygear.registry.Registry.get_provider')
    def testCallProviderContext(self, mocker):
        def assertState(expectedState):
            def func():
                assert current_context().get('state') == expectedState
            return func

        mocker.return_value = MagicMock(side_effect=assertState("happy"))
        self.transport.call_provider(self.ctx, 'name', 'action', {})
        mocker.return_value.handle_action.assert_called_once_with('action', {})
        assert 'state' not in current_context()

    @patch('skygear.registry.Registry.get_provider')
    def testCallProviderResult(self, mocker):
        provider_mock = MagicMock()
        provider_mock.handle_action.return_value = {'data': 'hello'}
        mocker.return_value = provider_mock
        result = self.transport.call_provider(self.ctx, 'name', 'action', {})
        assert result['result'] == {'data': 'hello'}

    @patch('skygear.registry.Registry.get_func')
    def testCallFuncGenericException(self, mocker):
        exc = Exception('Error occurred')
        mocker.return_value = MagicMock(side_effect=exc)
        result = self.transport.call_func(self.ctx, 'timer', 'name', {})
        assert result['error']['message'] == 'Error occurred'

    @patch('skygear.registry.Registry.get_func')
    def testCallFuncSkygearException(self, mocker):
        exc = SkygearException('Error occurred', 1, {'data': 'hello'})
        mocker.return_value = MagicMock(side_effect=exc)
        result = self.transport.call_func(self.ctx, 'timer', 'name', {})
        assert result['error']['message'] == 'Error occurred'
        assert result['error']['code'] == 1

    @patch('skygear.registry.Registry.get_event_funcs')
    def testCallEventFunc(self, mocker):
        mock1 = MagicMock()
        mock2 = MagicMock()
        mocker.return_value = [mock1, mock2]
        self.transport.call_event_func('some-event', {})
        mocker.assert_called_once_with('some-event')
        mock1.assert_called_once_with()
        mock2.assert_called_once_with()

    def testOpDictArg(self):
        mock = MagicMock(return_value={'result': 'OK'})
        self.transport.op(mock, dict(named='value'))
        mock.assert_called_with(named='value')

    def testOpArrayArg(self):
        mock = MagicMock(return_value={'result': 'OK'})
        self.transport.op(mock, [1, 2])
        mock.assert_called_with(1, 2)

    def testHandlerWithStrReturn(self):
        mock = MagicMock(return_value='Hello')
        response = self.transport.handler(mock, {
            'path': '/',
            'method': 'POST',
            'header': {'Content-Type': ['text/plain; charset=utf-8']},
            'body': base64.b64encode(b'rawstream')
        })
        assert response['header']['Content-Type'] == \
            ['text/plain; charset=utf-8']
        assert response['status'] == 200
        assert base64.b64decode(response['body']) == b'Hello'

    def testHandlerWithDictReturn(self):
        mock = MagicMock(return_value={
            'hello': 'world'
        })
        response = self.transport.handler(mock, {
            'path': '/',
            'method': 'POST',
            'header': {'Content-Type': ['text/plain; charset=utf-8']},
            'body': base64.b64encode(b'rawstream')
        })
        assert response['header']['Content-Type'] == ['application/json']
        assert response['status'] == 200
        assert base64.b64decode(response['body']) == b'{"hello": "world"}'

    def testHandlerWithResponseReturn(self):
        from werkzeug.utils import redirect

        mock = MagicMock(return_value=redirect('https://skygear.io/'))
        response = self.transport.handler(mock, {
            'path': '/',
            'method': 'POST',
            'header': {'Content-Type': ['text/plain; charset=utf-8']},
            'body': base64.b64encode(b'rawstream')
        })
        assert response['header']['Location'] == ['https://skygear.io/']
        assert response['status'] == 302

    def testHook(self):
        record_id = RecordID('note', 'note1')
        updated_record = Record(record_id, 'owner', None,
                                data={'data': 'updated'})
        old_record = Record(record_id, 'owner', None, data={'data': 'old'})
        new_record = Record(record_id, 'owner', None, data={'data': 'new'})
        param = {
            'original': serialize_record(old_record),
            'record': serialize_record(new_record),
            }
        mock = MagicMock(return_value=updated_record)
        returned_record = self.transport.hook(mock, param)
        assert mock.call_count == 1
        args, kwargs = mock.call_args
        assert args[0].data == new_record.data
        assert args[1].data == old_record.data
        assert deserialize_or_none(returned_record).data == updated_record.data

    def testTimer(self):
        mock = MagicMock()
        self.transport.timer(mock)
        mock.assert_called_once_with()

    def testProvider(self):
        mock = MagicMock()
        self.transport.provider(mock, 'action', {'data': 'hello'})
        mock.handle_action.assert_called_with('action', {'data': 'hello'})

    def testEvent(self):
        mock = MagicMock()
        self.transport.event(mock, {})
        mock.assert_called_once_with()


class TestBase64Encoding(unittest.TestCase):
    @patch.dict(os.environ, {'SKYGEAR_CONTEXT': 'e30='})
    def testFromEnviron(self):
        d = common.dict_from_base64_environ('SKYGEAR_CONTEXT')
        assert d == {}

    def testEncoding(self):
        assert common.encode_base64_json({}) == b'e30='

    def testDecoding(self):
        assert common.decode_base64_json('e30=') == {}


class TestHandleException(unittest.TestCase):
    def testHandleException(self):
        exc = Exception()
        assert common.handle_exception(exc) is exc

    def testHandleSkygearException(self):
        exc = SkygearException('Error occurred', 1, {'data': 'hello'})
        assert common.handle_exception(exc) is exc


class TestHandleInitEvent(unittest.TestCase):
    @patch('skygear.registry.Registry.func_list')
    def testHandleInitEvent(self, mocker):
        mocker.return_value = {'data': 'hello'}
        output = common.handle_init_event(hello='world')
        mocker.assert_called_once_with()
        assert output == mocker.return_value



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging

from pythonjsonlogger import jsonlogger

from .context import current_request_id


class CloudLogFormatter(jsonlogger.JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        super().add_fields(log_record, record, message_dict)
        log_record['level'] = record.levelname.lower()
        log_record['logger'] = record.name
        created = datetime.datetime.fromtimestamp(record.created)
        log_record['time'] = created.strftime('%Y-%m-%dT%H:%M:%SZ')

        if 'msg' not in log_record and 'message' in log_record:
            log_record['msg'] = log_record.pop('message')
        if 'error' not in log_record and 'exc_info' in log_record:
            log_record['error'] = log_record.pop('exc_info')


class RequestContextFilter(logging.Filter):
    def filter(self, record):
        request_id = current_request_id()
        if request_id:
            record.request_id = str(request_id)
        return True


class RequestTagFilter(logging.Filter):
    def __init__(self, tag, name=''):
        super(RequestTagFilter, self).__init__(name)
        self.tag = tag

    def filter(self, record):
        # We use the super class filter method to check if the filter name
        # applies to the record name. We modify the record if this is True.
        if super(RequestTagFilter, self).filter(record):
            if not hasattr(record, 'tag') and self.tag:
                record.tag = self.tag
        return True


def setLoggerTag(logger, tag):
    logger.addFilter(RequestTagFilter(tag))



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from collections import defaultdict
from copy import deepcopy

from .utils.logging import setLoggerTag

log = logging.getLogger(__name__)
setLoggerTag(log, 'plugin')


def get_registry():
    return _registry


def _iter_class_hierarchy(klass):
    klasses = [klass]
    while klasses:
        cls = klasses.pop()
        yield cls
        klasses += list(cls.__bases__)


class Registry:
    """
    Registry holds a mapping of registred functions and their parameters that
    are callable by Skygear plugin system.

    Developers are not expected to create an instance directly. Instead, they
    should use the function get_registry() to the shared instance.
    """

    def __init__(self):
        self.func_map = {
            'op': {},
            'hook': {},
            'timer': {},
        }
        self.event_map = defaultdict(list)
        self.param_map = {
            'op': [],
            'handler': [],
            'hook': [],
            'timer': [],
            'event': [],
            'provider': [],
        }
        self.handler = defaultdict(dict)
        self.providers = {}
        self.static_assets = {}
        self.exception_handlers = {}

    def _add_param_handler(self, param):
        new_param_list = []

        # Check existing param. Remove existing param if it overlaps with
        # the one being registered.
        for old_param in self.param_map['handler']:
            if not old_param.get('name') == param.get('name'):
                new_param_list.append(old_param)
                continue

            methods_intersection = set(old_param.get('methods')) & \
                set(param.get('methods'))

            if not methods_intersection:
                new_param_list.append(old_param)

        for method in param.get('methods'):
            new_param = deepcopy(param)
            new_param['methods'] = [method]
            new_param_list.append(new_param)
        self.param_map['handler'] = new_param_list

    def _add_param(self, kind, param):
        """
        Add a extension point param dict to the registry. If existing param
        dict overlaps the specified param dict, the existing param
        dict will be removed from the registry.

        An extension point param dict contains registration information of
        an extension point (i.e. cloud function).

        A param dict is an overlap of another when:
        - for handler: having the same name and method
        - others: having the same name
        """
        if kind == 'handler':
            return self._add_param_handler(param)

        param_list = self.param_map[kind]

        # Check existing param. Remove existing param if it overlaps with
        # the one being registered.
        for i in range(len(param_list)):
            if param_list[i].get('name') == param.get('name'):
                del param_list[i]
                break
        param_list.append(param)

    def register_hook(self, name, func, *args, **kwargs):
        if kwargs['type'] is None:
            raise ValueError("type is required for hook")
        if kwargs['trigger'] is None:
            raise ValueError("trigger is required for hook")
        kwargs['name'] = name

        if 'async_' in kwargs:
            kwargs['async'] = kwargs.pop('async_')

        if name in self.func_map['hook']:
            log.warning("Replacing previously registered hook '%s'.", name)

        self.func_map['hook'][name] = func
        self._add_param('hook', kwargs)

        log.debug("Registered hook '%s' to skygear!", name)

    def register_event(self, name, func, *args, **kwargs):
        self.event_map[name].append(func)

        self._add_param('event', {
            'name': name
        })

    def register_op(self, name, func, *args, **kwargs):
        if name in self.func_map['op']:
            log.warning("Replacing previously registered lambda '%s'.", name)

        self.func_map['op'][name] = func
        self._add_param('op', {
            'name': name,
            'key_required': kwargs.get('key_required',
                                       kwargs.get('auth_required', False)),
            'user_required': kwargs.get('user_required', False),
        })

        log.debug("Registered lambda '%s' to skygear!", name)

    def register_timer(self, name, func, *args, **kwargs):
        kwargs['name'] = name

        if name in self.func_map['op']:
            log.warning("Replacing previously registered timer '%s'.", name)

        self.func_map['timer'][name] = func
        self._add_param('timer', kwargs)

        log.debug("Registered timer '%s' to skygear!", name)

    def register_handler(self, name, func, *args, **kwargs):
        methods = kwargs.get('method', ['GET', 'POST', 'PUT'])
        if isinstance(methods, str):
            methods = [methods]

        for m in methods:
            if m in self.handler[name]:
                log.warning(
                    "Replacing previously registered handler '%s' (%s).",
                    name,
                    m
                )
            self.handler[name][m] = func
        self._add_param('handler', {
            'name': name,
            'methods': methods,
            'key_required': kwargs.get('key_required', False),
            'user_required': kwargs.get('user_required', False),
        })

        log.debug("Registered handler '%s' (%s) to skygear!",
                  name,
                  ', '.join(methods))

    def register_provider(self, provider_type, provider_id, provider,
                          **kwargs):
        kwargs['type'] = provider_type
        kwargs['id'] = provider_id

        if kwargs.get('name') is None:
            kwargs['name'] = provider_id

        if provider_id in self.providers:
            log.warning("Replacing previously registered provider '%s'.",
                        provider_id)

        self.providers[provider_id] = provider
        self._add_param('provider', kwargs)

        log.debug("Registered provider '%s' to skygear!", provider_id)

    def register_static_assets(self, prefix, func):
        self.static_assets[prefix] = func

    def get_static_assets(self, request_path):
        for prefix, loader in self.static_assets.items():
            if request_path.startswith(prefix):
                subpath = request_path[len(prefix):]
                return loader(), subpath
        raise KeyError('Unable to find static assets loader with '
                       'request_path "{}"'.format(request_path))

    def func_list(self):
        return self.param_map

    def get_func(self, kind, name):
        return self.func_map[kind][name]

    def get_event_funcs(self, name):
        return self.event_map[name]

    def get_provider(self, name):
        return self.providers[name]

    def get_handler(self, name, method):
        return self.handler[name].get(method, None)

    def get_exception_handler(self, klass):
        for cls in _iter_class_hierarchy(klass):
            if cls in self.exception_handlers:
                return self.exception_handlers[cls]
        return None

    def register_exception_handler(self, klass, func):
        self.exception_handlers[klass] = func


_registry = Registry()



# Copyright 2017 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .predicate import Predicate


class Query:
    """ Skygear Query Class

    Example:

        >>> p = Predicate(gender__eq="m")
        >>> query = Query("student", predicate=p)
        >>> database = container.public_database
        >>> result = database.query(query)

    """
    def __init__(self, record_type,
                 predicate=None, count=False,
                 limit=50, offset=None, include=[]):
        self.record_type = record_type
        if predicate is None:
            predicate = Predicate()
        self.predicate = predicate
        self.count = count
        self.sort = []
        self.limit = limit
        self.offset = offset
        self.include = include

    def add_order(self, key, order):
        self.sort.append([{'$type': 'keypath', '$val': key}, order])
        return self



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings

from skygear.encoding import (_RecordDecoder, _RecordEncoder,
                              deserialize_or_none, deserialize_record,
                              serialize_record)

# This file is to re-export the original package to maintain backward
# compability
warnings.warn("""Please import skygear.encoding instead,
skygear.transmitter.encoding is deperated""", DeprecationWarning)

__all__ = [
    'deserialize_record',
    'deserialize_or_none',
    'serialize_record',
    '_RecordDecoder',
    '_RecordEncoder',
]



import glob
import sys
from os import chdir, environ, mkdir, path, system
from shutil import copyfile, rmtree

from setuptools import Command, find_packages, setup
from setuptools.command.test import test as TestCommand

README = path.abspath(path.join(path.dirname(__file__), 'README.md'))

classifiers = [
    'License :: OSI Approved :: Apache Software License',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Operating System :: POSIX',
    'Operating System :: MacOS :: MacOS X',
    'Environment :: Web Environment',
    'Development Status :: 3 - Alpha',
]

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['skygear']
        self.test_suite = True
    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

class Doc(Command):
    """Sphinx Document Generation Command"""
    description = 'run sphinx-apidoc'
    user_options = []
    def finalize_options(self):
        pass

    def initialize_options(self):
        pass

    def run(self):
        metadata = self.distribution.metadata
        temp_dir = 'rst'
        chdir('docs')
        if path.exists(temp_dir) and path.isdir(temp_dir):
            rmtree(temp_dir)
        mkdir(temp_dir)
        environ['SPHINX_APIDOC_OPTIONS'] = 'members'
        system('sphinx-apidoc %s -H %s -A %s -V %s -R %s -o %s %s' %
               (path.join('..', 'skygear'),
                metadata.name,
                metadata.author,
                metadata.version,
                metadata.version,
                temp_dir,
                path.join('..', 'skygear')))
        for f in glob.glob(r'*.rst') + ['conf.py']:
            print(f)
            copyfile(f, path.join(temp_dir, f))
        system('sphinx-build %s %s' % (temp_dir, '_build'))
        rmtree(temp_dir)

extras_require={
    'zmq': ['pyzmq>=14.7'],
    'doc': ['Sphinx==1.6.7',
            'sphinx-rtd-theme>=0.2.4',
            'sphinxcontrib-napoleon>=0.6.1',
            'sphinxcontrib-websupport>=1.0.1']
}

setup(
      name='skygear',
      version='1.7.1',
      packages=find_packages(),
      description='Python plugin runtime for Skygear',
      long_description=open(README).read(),
      long_description_content_type='text/markdown',
      classifiers=classifiers,
      author='Rick Mak',
      author_email='rick.mak@gmail.com',
      url='https://github.com/SkygearIO/py-skygear',
      license='Apache License, Version 2.0',
      install_requires=[
            'psycopg2>=2.6.1,<3',
            'SQLAlchemy>=1.0.8,<2',
            'strict-rfc3339==0.5',
            'requests>=2.20.0,<3',
            'websocket-client==0.32.0',
            'bcrypt>=3.2.0,<4',
            'ConfigArgParse==0.12.0',
            'werkzeug>=0.11,<0.12',
            'boto3>=1.4,<2',
            'python-json-logger>=0.1.11,<0.2',
      ],
      extras_require=extras_require,
      cmdclass= {'test': PyTest, 'doc': Doc},
      tests_require=[
            'pytest',
      ] + extras_require['zmq']
        + extras_require['doc'],
      entry_points={
          'console_scripts': [
              'py-skygear = skygear.bin:main'
          ]
      },
)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from ..registry import Registry


class TestRegistry(unittest.TestCase):
    def test_register_handler(self):
        def handler():
            pass

        kwargs = {
            'method': ['GET', 'POST'],
            'key_required': True,
            'user_required': True,
        }
        registry = Registry()
        registry.register_handler('plugin:handler', handler, **kwargs)

        assert len(registry.handler) == 1
        assert registry.get_handler('plugin:handler', 'GET') == handler
        assert registry.get_handler('plugin:handler', 'POST') == handler
        assert registry.get_handler('plugin:handler', 'PUT') is None

        param_map = registry.param_map['handler']
        assert len(param_map) == 2
        assert param_map[0]['name'] == 'plugin:handler'
        assert 'GET' in param_map[0]['methods']
        assert param_map[0]['key_required'] is True
        assert param_map[0]['user_required'] is True
        assert param_map[1]['name'] == 'plugin:handler'
        assert 'POST' in param_map[1]['methods']
        assert param_map[1]['key_required'] is True
        assert param_map[1]['user_required'] is True

    def test_register_handler_with_different_method(self):
        def handler1():
            pass

        def handler2():
            pass

        def handler3():
            pass

        registry = Registry()
        registry.register_handler('plugin:handler', handler1,
                                  method=['GET', 'POST'])
        registry.register_handler('plugin:handler', handler2, method='POST')
        registry.register_handler('plugin:handler3', handler3, method='POST')

        assert len(registry.handler) == 2
        assert registry.get_handler('plugin:handler', 'GET') == handler1
        assert registry.get_handler('plugin:handler', 'POST') == handler2
        assert registry.get_handler('plugin:handler', 'PUT') is None
        assert registry.get_handler('plugin:handler3', 'POST') == handler3

        param_map = registry.param_map['handler']
        assert len(param_map) == 3
        assert param_map[0]['name'] == 'plugin:handler'
        assert param_map[0]['methods'] == ['GET']
        assert param_map[1]['name'] == 'plugin:handler'
        assert param_map[1]['methods'] == ['POST']
        assert param_map[2]['name'] == 'plugin:handler3'
        assert param_map[2]['methods'] == ['POST']

    def test_register_handler_twice(self):
        def handler1():
            pass

        def handler2():
            pass

        kwargs = {
            'method': ['GET', 'POST'],
            'key_required': True,
            'user_required': True,
        }
        registry = Registry()
        registry.register_handler('plugin:handler', handler1, **kwargs)
        registry.register_handler('plugin:handler', handler2, **kwargs)

        assert len(registry.handler) == 1
        assert registry.get_handler('plugin:handler', 'GET') == handler2
        assert registry.get_handler('plugin:handler', 'POST') == handler2
        assert registry.get_handler('plugin:handler', 'PUT') is None

        param_map = registry.param_map['handler']
        assert len(param_map) == 2
        assert param_map[0]['name'] == 'plugin:handler'
        assert 'GET' in param_map[0]['methods']
        assert param_map[0]['key_required'] is True
        assert param_map[0]['user_required'] is True

    def test_register_handler_with_one_str_method(self):
        def handler():
            pass

        kwargs = {
            'method': 'PUT',
        }
        registry = Registry()
        registry.register_handler('plugin:handler', handler, **kwargs)

        assert len(registry.handler) == 1
        assert registry.get_handler('plugin:handler', 'GET') is None
        assert registry.get_handler('plugin:handler', 'POST') is None
        assert registry.get_handler('plugin:handler', 'PUT') == handler

        param_map = registry.param_map['handler']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'plugin:handler'
        assert 'PUT' in param_map[0]['methods']
        assert param_map[0]['key_required'] is False
        assert param_map[0]['user_required'] is False

    def test_register_op(self):
        def fn():
            pass

        kwargs = {
                'key_required': True,
                'user_required': True,
                }
        registry = Registry()
        registry.register_op('plugin:action', fn, **kwargs)

        func_map = registry.func_map['op']
        assert len(func_map) == 1
        assert func_map['plugin:action'] == fn

        param_map = registry.param_map['op']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'plugin:action'
        assert param_map[0]['key_required'] is True
        assert param_map[0]['user_required'] is True

    def test_register_event(self):
        def fn():
            pass

        registry = Registry()
        registry.register_event('plugin:event:foo', fn)

        func_map = registry.event_map
        assert len(func_map) == 1
        assert func_map['plugin:event:foo'] == [fn]

        param_map = registry.param_map['event']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'plugin:event:foo'

    def test_register_hook(self):
        def fn():
            pass

        kwargs = {
                'type': 'note',
                'trigger': 'beforeSave',
                'async_': True,
                }
        registry = Registry()
        registry.register_hook('hook_name', fn, **kwargs)

        func_map = registry.func_map['hook']
        assert len(func_map) == 1
        assert func_map['hook_name'] == fn

        param_map = registry.param_map['hook']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'hook_name'
        assert param_map[0]['type'] == 'note'
        assert param_map[0]['trigger'] == 'beforeSave'
        assert param_map[0]['async'] is True

    def test_register_hook_twice(self):
        def fn1():
            pass

        def fn2():
            pass

        kwargs = {
                'type': 'note',
                'trigger': 'beforeSave',
                }
        registry = Registry()
        registry.register_hook('hook_name', fn1, **kwargs)
        registry.register_hook('hook_name', fn2, **kwargs)

        func_map = registry.func_map['hook']
        assert len(func_map) == 1
        assert func_map['hook_name'] == fn2

        param_map = registry.param_map['hook']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'hook_name'
        assert param_map[0]['type'] == 'note'
        assert param_map[0]['trigger'] == 'beforeSave'

    def test_register_timer(self):
        def fn():
            pass

        registry = Registry()
        registry.register_timer('timer_name', fn)

        func_map = registry.func_map['timer']
        assert len(func_map) == 1
        assert func_map['timer_name'] == fn

        param_map = registry.param_map['timer']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'timer_name'

    def test_register_timer_twice(self):
        def fn1():
            pass

        def fn2():
            pass

        registry = Registry()
        registry.register_timer('timer_name', fn1)
        registry.register_timer('timer_name', fn2)

        func_map = registry.func_map['timer']
        assert len(func_map) == 1
        assert func_map['timer_name'] == fn2

        param_map = registry.param_map['timer']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'timer_name'

    def test_register_provider(self):
        class Provider():
            pass

        the_provider = Provider()
        registry = Registry()
        registry.register_provider('example', 'com.example', the_provider)

        assert len(registry.providers) == 1
        assert registry.providers['com.example'] == the_provider

        param_map = registry.param_map['provider']
        assert len(param_map) == 1
        assert param_map[0]['type'] == 'example'
        assert param_map[0]['id'] == 'com.example'
        assert param_map[0]['name'] == 'com.example'

    def test_register_multiple_providers(self):
        class Provider():
            pass

        provider1 = Provider()
        provider2 = Provider()
        registry = Registry()
        registry.register_provider('example', 'com.example.1', provider1)
        registry.register_provider('example', 'com.example.2', provider2)

        assert len(registry.providers) == 2
        assert registry.providers['com.example.1'] == provider1
        assert registry.providers['com.example.2'] == provider2

        param_map = registry.param_map['provider']
        assert len(param_map) == 2
        assert param_map[0]['type'] == 'example'
        assert param_map[0]['id'] == 'com.example.1'
        assert param_map[0]['name'] == 'com.example.1'
        assert param_map[1]['type'] == 'example'
        assert param_map[1]['id'] == 'com.example.2'
        assert param_map[1]['name'] == 'com.example.2'

    def test_register_static_assets(self):
        def fn():
            pass

        registry = Registry()
        registry.register_static_assets('admin', fn)

        assert len(registry.static_assets) == 1
        assert registry.static_assets['admin'] == fn

    def test_get_static_assets(self):
        class Loader:
            pass

        loader = Loader()

        def fn():
            return loader

        registry = Registry()
        registry.static_assets['admin'] = fn
        got_loader, subpath = registry.get_static_assets('admin/apple/pie')

        assert got_loader is loader
        assert subpath == '/apple/pie'

    def test_get_static_assets_with_traling(self):
        class Loader:
            pass

        loader = Loader()

        def fn():
            return loader

        registry = Registry()
        registry.static_assets['admin/'] = fn
        got_loader, subpath = registry.get_static_assets('admin/apple/pie')

        assert got_loader is loader
        assert subpath == 'apple/pie'

    def test_register_exception_handler(self):
        def fn(exc):
            pass

        registry = Registry()
        registry.register_exception_handler(Exception, fn)

        assert len(registry.exception_handlers) == 1
        assert registry.exception_handlers[Exception] == fn

    def test_get_exception_handler(self):
        class SomeException(Exception):
            pass

        def fn(exc):
            return exc

        registry = Registry()
        registry.exception_handlers[Exception] = fn
        handler = registry.get_exception_handler(SomeException)

        assert handler is fn

    def test_register_lambda(self):
        def fn():
            return True

        registry = Registry()
        registry.register_op('plugin:action', fn, key_required=True)
        assert registry.func_map['op']['plugin:action'] == fn
        assert registry.param_map['op'][0]['key_required'] is True

    def test_register_lambda_twice(self):
        def fn1():
            return True

        def fn2():
            return False

        registry = Registry()
        registry.register_op('plugin:action', fn1, key_required=True)
        registry.register_op('plugin:action', fn2, key_required=False)

        assert len(registry.func_map['op']) == 1
        assert registry.func_map['op']['plugin:action'] == fn2
        assert len(registry.param_map['op']) == 1
        assert registry.param_map['op'][0]['key_required'] is False



# Copyright 2017 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
import copy


class Predicate(object):
    """The Skygear predicate class.

    Example:

    >>> p = Predicate(_id__in=["a", "b", "c"])
    >>> p = Predicate(_id__eq="simple", gender__eq="M") #Conjuction
    >>> p = Predicate(_id__eq="simple", gender__eq="M", op=Predicate.OR) #Disjunction
    >>> p = ~Predicate(_id__eq="milktea") #Negation
    """ # noqa

    AND = 'and'
    OR = 'or'
    NOT = 'not'

    def __init__(self, **kwargs):
        self.op = kwargs.pop('op', Predicate.AND)
        # Need to use ordered dict in order to pass unit test in python 3.5.
        # See https://docs.python.org/3.6/whatsnew/3.6.html#whatsnew36-pep468
        od = collections.OrderedDict(sorted(kwargs.items()))
        self.conditions = [(key, kwargs[key]) for key in od.keys()]

    def __and__(self, other):
        new_instance = Predicate()
        if self.op == Predicate.AND:
            new_instance.conditions = copy.copy(self.conditions)
            if other.op == self.op:
                new_instance.conditions += other.conditions
            else:
                new_instance.conditions.append(other)
        else:
            new_instance.conditions = [self, other]
        return new_instance

    def __or__(self, other):
        new_instance = Predicate(op=Predicate.OR)
        if self.op == Predicate.OR:
            new_instance.conditions = copy.copy(self.conditions)
            if other.op == self.op:
                new_instance.conditions += other.conditions
            else:
                new_instance.conditions.append(other)
        else:
            new_instance.conditions = [self, other]
        return new_instance

    def __invert__(self):
        new_instance = Predicate(op=Predicate.NOT)
        new_instance.conditions = [self]
        return new_instance

    @classmethod
    def condition_to_dict(cls, t):
        field, op = t[0].split("__")
        return [op, {"$type": "keypath", "$val": field}, t[1]]

    def to_dict(self, root=None):
        if root is None:
            root = self
        if root is None:
            return []
        if isinstance(root, Predicate):
            num_conditions = len(root.conditions)
            if num_conditions == 0:
                return []
            elif num_conditions == 1:
                result = self.to_dict(root.conditions[0])
                if root.op != Predicate.NOT:
                    return result
                return [root.op, result]
            else:
                return [root.op] + [self.to_dict(d) for d in root.conditions]
        elif type(root) == tuple:
            return Predicate.condition_to_dict(root)
        else:
            return []
